
//---------------------------------------------------------
/**    @file		Surface.cpp
*     @brief	Surface.cpp is the Surface CLASS
*															*/
//---------------------------------------------------------

#include "Surface.h"

Surface::Surface()
{
	delphi = NULL;
	panel = 0;
	totalSurfaceArea=0;
	totalVolume=0;	
	inside = 1;
	accurateTriangulation = true;
	isAvailableScalarField = false;
	projBGP = true;
	scalarField = NULL;
	delta_accurate_triangulation = 1e-10;
	checkDuplicatedVertices = false;
	wellShaped = false;
	probe_radius = 1.4;
	last_rows_ind = -1;
	last_cols_ind = -1;
}

Surface::~Surface()
{	
	if (triList.size()>0)
	{
		vector<int*>::iterator it;
		for (it = triList.begin();it!=triList.end();it++)
			deleteVector<int>((*it));
	}

	if (vertList.size()>0)
	{
		vector<double*>::iterator it;
		for (it = vertList.begin();it!=vertList.end();it++)
			deleteVector<double>((*it));
	}

	if (intersectionsMatrixAlongX!=NULL)
	{
		for (int i=0;i<delphi->nx;i++)
			for (int j=0;j<delphi->ny;j++)
				for (int k=0;k<delphi->nx;k++)
					if (intersectionsMatrixAlongX[i][j][k]!=NULL)
						deleteVector<double>(intersectionsMatrixAlongX[i][j][k]);

		deleteMatrix3D<double*>(delphi->nx,delphi->ny,intersectionsMatrixAlongX);
	}

	if (intersectionsMatrixAlongY!=NULL)
	{
		for (int i=0;i<delphi->nx;i++)
			for (int j=0;j<delphi->ny;j++)
				for (int k=0;k<delphi->nx;k++)
					if (intersectionsMatrixAlongY[i][j][k]!=NULL)
						deleteVector<double>(intersectionsMatrixAlongY[i][j][k]);

		deleteMatrix3D<double*>(delphi->nx,delphi->ny,intersectionsMatrixAlongY);
	}

	if (intersectionsMatrixAlongZ!=NULL)
	{
		for (int i=0;i<delphi->nx;i++)
			for (int j=0;j<delphi->ny;j++)
				for (int k=0;k<delphi->nx;k++)
					if (intersectionsMatrixAlongZ[i][j][k]!=NULL)
						deleteVector<double>(intersectionsMatrixAlongZ[i][j][k]);

		deleteMatrix3D<double*>(delphi->nx,delphi->ny,intersectionsMatrixAlongZ);
	}
	
	if (verticesInsidenessMap!=NULL)
		deleteMatrix3D<bool>(delphi->nx,delphi->ny,verticesInsidenessMap);

	if (scalarField!=NULL)
		deleteMatrix3D<double>(delphi->nx,delphi->ny,scalarField);
}

bool Surface::build()
{
	cout << endl << WARN << "Build surface not supported!";
	return false;
}


bool Surface::save(char* fileName)
{
	cout << endl << WARN << "Save surface not supported!";
	return false;
}

bool Surface::load(char* fileName)
{
	cout << endl << WARN << "Load surface not supported";
	return false;
}

void Surface::printSummary()
{
	cout << endl << WARN << "Print summary not supported!";
}


/** build epsmap and compute boundary grid points by intersection
and projection routines. The minimal number of projections is performed.
The ray tracing part is parallelized with boost threading if enabled*/
bool Surface::getSurf(bool fillCav,double vol)
{
	if (delphi == NULL)
	{
		cout << endl << WARN << "Cannot get surface without DelPhi environment!";
		return false;
	}

	if (accurateTriangulation)
	{	
		intersectionsMatrixAlongX = allocateMatrix3D<double*>(delphi->nx,delphi->ny,delphi->nz); 
		intersectionsMatrixAlongY = allocateMatrix3D<double*>(delphi->nx,delphi->ny,delphi->nz); 
		intersectionsMatrixAlongZ = allocateMatrix3D<double*>(delphi->nx,delphi->ny,delphi->nz); 
		verticesInsidenessMap = allocateMatrix3D<bool>(delphi->nx,delphi->ny,delphi->nz); 
		for (int i=0;i<delphi->nx;i++)
			for (int j=0;j<delphi->ny;j++)
				for (int k=0;k<delphi->nz;k++)
				{
					intersectionsMatrixAlongX[i][j][k]=NULL;
					intersectionsMatrixAlongY[i][j][k]=NULL;
					intersectionsMatrixAlongZ[i][j][k]=NULL;
					verticesInsidenessMap[i][j][k]=true;
				}
	}
	else
	{
		intersectionsMatrixAlongX = NULL;
		intersectionsMatrixAlongY = NULL;
		intersectionsMatrixAlongZ = NULL;
		verticesInsidenessMap = NULL;
	}	

	int na,nb;
	double** volPanelTHD;
	double volPanel[3];

	cout << endl << INFO << "Inside id value is " << inside;
	
	#ifdef ENABLE_BOOST
		int num_thd;		
		if (num_cpus==-1)
		{
			num_thd = boost::thread::hardware_concurrency();
			cout << endl << INFO << "Detected " << num_thd << " cores";
		}
		else
		{
			cout << endl << INFO << "Forcing user selected num threads " << num_cpus;
			num_thd = num_cpus;
		}
	#else
		int num_thd = 1;
	#endif

	#ifdef DEBUG_SURFACE
		// force 1 thread for debug
		num_thd = 1;
	#endif

	volPanelTHD = allocateMatrix2D<double>(num_thd+1,3);
	int* numIntersections = allocateVector<int>(num_thd+1);
	
	cout << endl << INFO << "Using " << num_thd << " thread[s]";

	time_t start, end;
	double duration = 0;
	int numint = 0;
	time(&start);

	for (int l=0;l<num_thd+1;l++)
		numIntersections[l]=0;

	// Phase 1, ray trace from each coordinate plane	
	for (panel=0;panel<3;panel++)
	{		
		preProcessPanel();
		cout << endl << INFO << "Ray-tracing panel " << panel << "....";
		volPanel[panel]=0;
	
		// left y,z panel
		if (panel==0)
		{
			na = delphi->ny;
			nb = delphi->nz;
		}
		// lower y,x panel
		else if (panel==1)
		{
			na = delphi->ny;
			nb = delphi->nx;
		}
		// back z,x panel
		else
		{
			na = delphi->nz;
			nb = delphi->nx;		
		}
								
		int cycle = na/num_thd;
		int rem = na%num_thd;
		
		//cout << endl << "cycle " << cycle;
		//cout << endl << "rem " << rem;

		#ifdef ENABLE_BOOST
			boost::thread_group thdGroup; 
		#endif
		
		// setup split
		int j=0;
		
		
		for (j=0;j<num_thd;j++)
		{
			int start = cycle*j;
			int stop = cycle*(j+1);
			#ifdef ENABLE_BOOST
				thdGroup.create_thread(boost::bind(&Surface::intersector,this,volPanelTHD[j],nb,start,stop,&(numIntersections[j]),j));
			#else
				intersector(volPanelTHD[j],nb,start,stop,&(numIntersections[j]),0);
			#endif
		}
		
		// remaining part
		if (rem>0)
		{
			int start = cycle*j;
			int stop = start+rem;
			#ifdef ENABLE_BOOST
				thdGroup.create_thread(boost::bind(&Surface::intersector,this,volPanelTHD[j],nb,start,stop,&(numIntersections[j]),j));
			#else
				intersector(volPanelTHD[j],nb,start,stop,&(numIntersections[j]),1);
			#endif
		}
		// end setup
		
		#ifdef ENABLE_BOOST
			// join
			thdGroup.join_all();
			// reduce
			for (j=0;j<num_thd;j++)
			{	
				volPanel[panel]+=volPanelTHD[j][panel];
				numint+=numIntersections[j];
			}

			if (rem>0)
			{
				volPanel[panel]+=volPanelTHD[j][panel];
				numint+=numIntersections[j];
			}
			
		#else			
			volPanel[panel]=volPanelTHD[0][panel];
			numint = numIntersections[0];
		#endif
		
		cout << "ok!";		
	}                    		

	deleteMatrix2D<double>(num_thd+1,volPanelTHD);	

	numint = 0;
	// record all the approximated intersections
	for (int l=0;l<num_thd+1;l++)
		numint+=numIntersections[l];

	deleteVector<int>(numIntersections);
	time(&end);
	duration = difftime(end, start);
	cout << endl << INFO << "Ray-tracing computation time.. " << duration << " [s]"; 	
	// assuming squared grid for this stat
	if (accurateTriangulation && !isAvailableScalarField)
		printf("\n%sApproximated %d rays (%.5f %%)",INFO,numint,numint/(6*(float)delphi->nx*delphi->ny)*100);
	else
		printf("\n%sApproximated %d rays (%.5f %%)",INFO,numint,numint/(3*(float)delphi->nx*delphi->ny)*100);
	
	vector<int*> bgp;

	time_t start_cav,end_cav;
	time(&start_cav);
	// before bgp identification, cavities are filled if requested
	if (fillCav)
	{
		cout << endl << INFO << "Performing cavity detection and conditional filling...";
		int cav = getCavities();
		cout << "ok!";
		cout << endl << INFO << "Detected " << cav << " cavitiy[ies]";		
		fillCavities(vol);
		if (wellShaped)
		{
			cout << endl << INFO << "Performing cavities shape filtering..";
			filterCavities();
		}
	}
	time(&end_cav);
	duration = difftime(end_cav, start_cav);
	cout << endl << INFO << "Cavity detection time is " << duration <<  " [s]";
	
	// if only buildEpsMap is active, this correction due to cavity detection
	// is not necessary
	if (delphi->buildEpsMap && delphi->buildStatus)
	{
		cout << endl << INFO << "Writing idebmap...";
		// write idebmap
		for (int i=0;i<delphi->nx;i++)
			for (int j=0;j<delphi->ny;j++)
				for (int k=0;k<delphi->nz;k++)
					// inside
					if (delphi->status[i][k][j]==STAUTS_POINT_INSIDE)
						delphi->IDEBMAP(i,j,k,(delphi->nx),(delphi->ny))=false;
					// outside
					else
						delphi->IDEBMAP(i,j,k,(delphi->nx),(delphi->ny))=true;
		cout << "ok!";
	}

	int NX = delphi->nx;
	int NY = delphi->ny;
	int NZ = delphi->nz;

	// rest bgp count
	delphi->nbgp = 0;

	postRayCasting();

	// Phase 2. boundary grid points are located and projected if requested
	if (projBGP)
	{
		preBoundaryProjection();
		cout << endl << INFO << "Detecting boundary grid points...";
		for (int iz=0;iz<NZ;iz++)	
		{
			for (int iy=0;iy<NY;iy++)
			{
				for (int ix=0;ix<NX;ix++)
				{				
					int kost = delphi->EPSMAP(ix,iy,iz,0,NX,NY,NZ);
					
					//printf("\n delphi->EPSMAP(%d) = %d",(ix)+(NX)*((iy)+(NY)*((iz)+(NZ)*(0))),delphi->EPSMAP(ix,iy,iz,0,NX,NY,NZ));
					
					//char kost = delphi->epsmapx[ix][iy][iz];
					if ((ix-1<0) || (iy-1<0) || (iz-1<0))
						continue;

					if (kost != delphi->EPSMAP(ix,iy,iz,1,NX,NY,NZ) ||
					kost != delphi->EPSMAP(ix,iy,iz,2,NX,NY,NZ) ||
					kost != delphi->EPSMAP((ix-1),iy,iz,0,NX,NY,NZ) ||
					kost != delphi->EPSMAP(ix,(iy-1),iz,1,NX,NY,NZ) ||
					kost != delphi->EPSMAP(ix,iy,(iz-1),2,NX,NY,NZ) )
					{

						int* temp = allocateVector<int>(3);
						temp[0] = ix;
						temp[1] = iy;
						temp[2] = iz;

						double gridPoint[3];
						gridPoint[0] = delphi->x[ix];
						gridPoint[1] = delphi->y[iy];
						gridPoint[2] = delphi->z[iz];
						bgp.push_back(temp);
					}
	            
				}
			}
		}

		delphi->nbgp = bgp.size();
		if (!delphi->getDelphiBinding())
		{
			delphi->scspos	= allocateVector<double>(3*delphi->nbgp);
			delphi->scsnor	= allocateVector<double>(3*delphi->nbgp);
			delphi->scsarea = allocateVector<double>(delphi->nbgp);
			delphi->ibgp	= allocateVector<int>(3*delphi->nbgp);
		}


		if (delphi->getDelphiBinding() && bgp.size()>((unsigned)delphi->maxbgp))
		{
			cout << endl << ERR << "Number of bgp is " << bgp.size() << " and the maximum allowed is " << delphi->maxbgp;
			cout << endl << ERR << "Please increase ibmx in DelPhi and recompile";
			exit(-1);
		}

		// freeze vector
		vector<int*>::iterator it;
		int i=0;
		for (it=bgp.begin();it!=bgp.end();it++)
		{
			//delphi->ibgp[i] = *it;
			int* vv = *it;
			delphi->ibgp[3*i]=vv[0];
			delphi->ibgp[3*i+1]=vv[1];
			delphi->ibgp[3*i+2]=vv[2];
			deleteVector<int>(vv);
			i++;
		}
		cout << "ok!";
		cout << endl << INFO << "Detected " << delphi->nbgp << " boundary grid points (bgp)";
		cout << endl << INFO << "Scaling bgps...";

		// single threaded
		/*for (i=0;i<delphi->nbgp;i++)
		{
			double gridPoint[3];
			gridPoint[0] = delphi->x[delphi->ibgp[3*i]];
			gridPoint[1] = delphi->y[delphi->ibgp[3*i+1]];
			gridPoint[2] = delphi->z[delphi->ibgp[3*i+2]];
			getProjection(gridPoint,&(delphi->scspos[3*i]),
									&(delphi->scspos[3*i+1]),
									&(delphi->scspos[3*i+2]),
									&(delphi->scsnor[3*i]),
									&(delphi->scsnor[3*i+1]),
									&(delphi->scsnor[3*i+2]));
		}*/

		int cycle = (delphi->nbgp)/num_thd;
		int rem = (delphi->nbgp)%num_thd;
		//cout << endl << INFO << "Block size " << cycle << " Rem " << rem;

		#ifdef ENABLE_BOOST
			boost::thread_group thdGroup; 
		#endif
		
		// setup split
		int j=0;
		
		for (j=0;j<num_thd;j++)
		{
			int start = cycle*j;
			int stop = cycle*(j+1);
			#ifdef ENABLE_BOOST
				thdGroup.create_thread(boost::bind(&Surface::projector,this,start,stop));
			#else
				projector(start,stop);
			#endif
		}
		
		// remaining part
		if (rem>0)
		{
			int start = cycle*j;
			int stop = start+rem;
			#ifdef ENABLE_BOOST
				thdGroup.create_thread(boost::bind(&Surface::projector,this,start,stop));
			#else
				projector(start,stop);
			#endif
		}
		// end setup
		
		#ifdef ENABLE_BOOST
			// join
			thdGroup.join_all();
		#endif

		cout << "ok!";		
	}                    		
	
	volPanel[0]*=delphi->A;		
	volPanel[1]*=delphi->A;
	volPanel[2]*=delphi->A;

	//cout << endl << "Vol 1 " << volPanel[0] << " Vol 2 " << volPanel[1] <<  " Vol 3 " << volPanel[2];

	// compute the total volume as the average of the 3 volumes

	if (delphi->buildStatus && !delphi->buildEpsMap && panel!=0) 
		totalVolume = volPanel[0];
	else
		totalVolume=(volPanel[0]+volPanel[1]+volPanel[2])/3;

	return true;
}


double Surface::getVolume()
{
	return totalVolume;
}


int Surface::getCavities()
{
	// search cavities in status vector
	bool cavities=true;
	int id = STATUS_POINT_TEMPORARY_OUT;

	// free memory 
	if (delphi->cavitiesVec !=NULL)
	{
		vector< vector < int* >* >::iterator it;
		for (it=delphi->cavitiesVec->begin();it!=delphi->cavitiesVec->end();it++)
		{
			vector < int* >* inner = (*it);
			vector < int* >::iterator it2;
			for (it2=inner->begin();it2!=inner->end();it2++)
				free((*it2));
			delete inner;
		}
	}
	delete delphi->cavitiesVec;

	// allocate empty vector
	delphi->cavitiesVec = new vector< vector < int* >* >();
	
	while(1)
	{
		int i,j,k;
		cavities = false;
		// get starting point
		for (i=0;i<delphi->nx;i++)
		{
			for (j=0;j<delphi->ny;j++)
			{
				for (k=0;k<delphi->nz;k++)
				{
					if (delphi->status[i][j][k]==STATUS_POINT_TEMPORARY_OUT)
					{
						cavities = true;
						break;
					}
				}
				if (cavities)
					break;
			}
			if (cavities)
				break;
		}

		if (!cavities)
			break;

		// new status
		id++;

		// mark from temporary outside to cavity index.
		// the first pass is from STATUS_POINT_TEMPORARY_OUT -> STATUS_POINT_OUT this is not cavity, this is only real outside
		// from that moment on, if there are still cavities they are all marked with STATUS_POINT_TEMPORARY_OUT
		// and one by one they are detected and associated to a cavity index
		floodFill(i,j,k,STATUS_POINT_TEMPORARY_OUT,id);	
	}

	int numCavities = id-STATUS_POINT_OUT;	
	delphi->cavitiesSize.clear();	
	delphi->cavitiesSize.resize(numCavities);
	delphi->cavitiesFlag.resize(numCavities);
	return numCavities;
}

void Surface::fillCavities(double vol)
{

	if (vol<0)
	{
		cout << endl << WARN << "Cannot fill with a negative volume. Setting " << DEFAULT_VOLUME;
		vol = DEFAULT_VOLUME;
	}
	// analyze each cavity and fill if required
	vector < vector <int* >* >::iterator it;
	int i=0;
	double cubeVol = delphi->side*delphi->side*delphi->side;

	int NX = delphi->nx;
	int NY = delphi->ny;
	int NZ = delphi->nz;

	cout << endl << INFO << "Threshold volume is " << vol;
	for (it=delphi->cavitiesVec->begin();it!=delphi->cavitiesVec->end();it++)
	{
		cout << endl << INFO << "Cavity " << i ;
		double cavVol = (*it)->size()*cubeVol;
		printf("\tvol is %.4lf [A^3] \t",cavVol);
		delphi->cavitiesSize[i]=cavVol;
		delphi->cavitiesFlag[i]=false;

		if (cavVol<=vol)
		{
			delphi->cavitiesFlag[i]=true;
			vector <int* >::iterator it2;
			vector<int*>* vec = (*it);
			// filling eps map
			for (it2=vec->begin();it2!=vec->end();it2++)
			{
				int* v = (*it2);
				// that grid point is filled
				/*delphi->epsmapx[v[0]][v[1]][v[2]]=-1;
				delphi->epsmapy[v[0]][v[1]][v[2]]=-1;
				delphi->epsmapz[v[0]][v[1]][v[2]]=-1;
				delphi->epsmapx[v[0]-1][v[1]][v[2]]=-1;
				delphi->epsmapy[v[0]][v[1]-1][v[2]]=-1;
				delphi->epsmapz[v[0]][v[1]][v[2]-1]=-1;*/

				// apply correction only if epsmap is used
				if (delphi->buildEpsMap)
				{
					delphi->EPSMAP((v[0]),(v[1]),(v[2]),0,NX,NY,NZ)=inside;
					delphi->EPSMAP((v[0]),(v[1]),(v[2]),1,NX,NY,NZ)=inside;
					delphi->EPSMAP((v[0]),(v[1]),(v[2]),2,NX,NY,NZ)=inside;
					delphi->EPSMAP((v[0]-1),(v[1]),(v[2]),0,NX,NY,NZ)=inside;
					delphi->EPSMAP((v[0]),(v[1]-1),(v[2]),1,NX,NY,NZ)=inside;
					delphi->EPSMAP((v[0]),(v[1]),(v[2]-1),2,NX,NY,NZ)=inside;
					
					delphi->IDEBMAP((v[0]),(v[1]),(v[2]),NX,NY) = false;
				}

				if (accurateTriangulation && !isAvailableScalarField)
				{
					verticesInsidenessMap[v[0]][v[1]][v[2]] = false;
					verticesInsidenessMap[v[0]+1][v[1]][v[2]] = false;
					verticesInsidenessMap[v[0]][v[1]+1][v[2]] = false;
					verticesInsidenessMap[v[0]+1][v[1]+1][v[2]] = false;

					verticesInsidenessMap[v[0]][v[1]][v[2]+1] = false;
					verticesInsidenessMap[v[0]+1][v[1]][v[2]+1] = false;
					verticesInsidenessMap[v[0]][v[1]+1][v[2]+1] = false;
					verticesInsidenessMap[v[0]+1][v[1]+1][v[2]+1] = false;
				}
				// update the scalar filed consistently with cavity detection
				if (accurateTriangulation && isAvailableScalarField)
				{
					scalarField[v[0]][v[1]][v[2]] = +INFINITY;
					scalarField[v[0]+1][v[1]][v[2]] = +INFINITY;
					scalarField[v[0]][v[1]+1][v[2]] = +INFINITY;
					scalarField[v[0]+1][v[1]+1][v[2]] = +INFINITY;

					scalarField[v[0]][v[1]][v[2]+1] = +INFINITY;
					scalarField[v[0]+1][v[1]][v[2]+1] = +INFINITY;
					scalarField[v[0]][v[1]+1][v[2]+1] = +INFINITY;
					scalarField[v[0]+1][v[1]+1][v[2]+1] = +INFINITY;
				}
			}
			cout << "filled";
		}
		else
		{
			cout << "non filled";
		}
		i++;
	}
}

void Surface::filterCavities()
{	
	int ngrid = (int)rintp(2.*probe_radius/delphi->side);
	if ((ngrid%2)==0)
		ngrid++;

	int k = (ngrid-1)/2;
	cout << " Probe radius is " << probe_radius << " [A], grid points " << k;
	int cavityId;

	double dist2ref = probe_radius*probe_radius;
	vector < vector <int* >* >::iterator it;
	int i=0;

	int NX = delphi->nx;
	int NY = delphi->ny;
	int NZ = delphi->nz;

	// analyze each cavity
	for (it=delphi->cavitiesVec->begin();it!=delphi->cavitiesVec->end();it++,i++)
	{
		cout << endl << INFO << "Cavity " << i ;
		if (delphi->cavitiesFlag[i])
		{
			cout << " already filled";
			continue;
		}
		else
			cout << " Under analysis...";

		vector<int*>* currentCavity = (*it);
		vector <int* >::iterator it2;		

		// mark each cavity point as temporary STATUS_CAVITY_POINT_SHAPE_UNDER_CHECK code
		for (it2=currentCavity->begin();it2!=currentCavity->end();it2++)
		{
			int* v = *it2;
			cavityId = delphi->status[v[0]][v[1]][v[2]];
			delphi->status[v[0]][v[1]][v[2]] = STATUS_CAVITY_POINT_SHAPE_UNDER_CHECK;
		}

		// for each cavity point 'march' around it in order to understand if the 
		// water molecule can fit in the cavity
		
		// a cavity is not filled if at least once a water molecule can stay into the cavity.

		bool keepCavity = false;

		for (it2=currentCavity->begin();it2!=currentCavity->end();it2++)
		{
			int* v = *it2;			
			int start_ix = v[0]-k;
			int start_iy = v[1]-k;
			int start_iz = v[2]-k;

			int end_ix = v[0]+k;
			int end_iy = v[1]+k;
			int end_iz = v[2]+k;
		
			bool oneProbeFits = true;

			for (int ix = start_ix;ix<end_ix;ix++)
			{
				for (int iy = start_iy;iy<end_iy;iy++)
				{
					for (int iz = start_iz;iz<end_iz;iz++)
					{												
						double a = (delphi->x[ix]-delphi->x[v[0]])*(delphi->x[ix]-delphi->x[v[0]]);
						double b = (delphi->y[iy]-delphi->y[v[1]])*(delphi->y[iy]-delphi->y[v[1]]);
						double c = (delphi->z[iz]-delphi->z[v[2]])*(delphi->z[iz]-delphi->z[v[2]]);

						double dist2 = a+b+c;
						if (dist2>dist2ref)
							continue;

						// there is no space we went outside the cavity, go on to the next grid point
						if (delphi->status[ix][iy][iz]!=STATUS_CAVITY_POINT_SHAPE_UNDER_CHECK)
						{
							oneProbeFits = false;
							break;
						}
					}
					if (!oneProbeFits)
						break;
				}
				if (!oneProbeFits)
					break;
			}

			// water molecule fits, stop keeping cavity
			if (oneProbeFits)
			{
				keepCavity = true;
				break;
			}
		}

		for (it2=currentCavity->begin();it2!=currentCavity->end();it2++)
		{	
			int* v = *it2;								
			// clean cavity
			if (!keepCavity)
			{		

				if (delphi->buildEpsMap)
				{
					delphi->EPSMAP((v[0]),(v[1]),(v[2]),0,NX,NY,NZ)=inside;
					delphi->EPSMAP((v[0]),(v[1]),(v[2]),1,NX,NY,NZ)=inside;
					delphi->EPSMAP((v[0]),(v[1]),(v[2]),2,NX,NY,NZ)=inside;
					delphi->EPSMAP((v[0]-1),(v[1]),(v[2]),0,NX,NY,NZ)=inside;
					delphi->EPSMAP((v[0]),(v[1]-1),(v[2]),1,NX,NY,NZ)=inside;
					delphi->EPSMAP((v[0]),(v[1]),(v[2]-1),2,NX,NY,NZ)=inside;
					
					delphi->IDEBMAP((v[0]),(v[1]),(v[2]),NX,NY) = false;
				}

				if (accurateTriangulation && !isAvailableScalarField)
				{
					verticesInsidenessMap[v[0]][v[1]][v[2]] = false;
					verticesInsidenessMap[v[0]+1][v[1]][v[2]] = false;
					verticesInsidenessMap[v[0]][v[1]+1][v[2]] = false;
					verticesInsidenessMap[v[0]+1][v[1]+1][v[2]] = false;

					verticesInsidenessMap[v[0]][v[1]][v[2]+1] = false;
					verticesInsidenessMap[v[0]+1][v[1]][v[2]+1] = false;
					verticesInsidenessMap[v[0]][v[1]+1][v[2]+1] = false;
					verticesInsidenessMap[v[0]+1][v[1]+1][v[2]+1] = false;
				}
				
				// update the scalar filed consistently with cavity shape detection
				if (accurateTriangulation && isAvailableScalarField)
				{
					scalarField[v[0]][v[1]][v[2]] = +INFINITY;
					scalarField[v[0]+1][v[1]][v[2]] = +INFINITY;
					scalarField[v[0]][v[1]+1][v[2]] = +INFINITY;
					scalarField[v[0]+1][v[1]+1][v[2]] = +INFINITY;

					scalarField[v[0]][v[1]][v[2]+1] = +INFINITY;
					scalarField[v[0]+1][v[1]][v[2]+1] = +INFINITY;
					scalarField[v[0]][v[1]+1][v[2]+1] = +INFINITY;
					scalarField[v[0]+1][v[1]+1][v[2]+1] = +INFINITY;
				}
			}
			
			// restore orginal value in any case
			delphi->status[v[0]][v[1]][v[2]] = cavityId;
		} // end for each cavity point

		if (!keepCavity)	
		{
			delphi->cavitiesFlag[i] = true;					
			cout << "Filled";
		}
		else
			cout << "Non filled";

	} // end for each cavity	
}


/** queue based implementation*/
void Surface::floodFill(int ix,int iy,int iz,int idold,int idnew)
{
	queue<pair<int,pair<int,int> > > myqueue;
	pair<int,pair<int,int> > triplet;
	int cix,ciy,ciz;

	myqueue.push(pair<int,pair<int,int > >(ix,pair<int,int>(iy,iz)));

	while(myqueue.size()!=0)
	{	
		triplet = myqueue.front();
		myqueue.pop();
		
		cix = triplet.first;
		ciy = triplet.second.first;
		ciz = triplet.second.second;

		//cout << endl << myqueue.size();
		//cout << endl <<cix << " " << ciy << " " << ciz;
		if (delphi->status[cix][ciy][ciz]==idold)
		{
			delphi->status[cix][ciy][ciz]=idnew;

			// if this is a cavity then update the cavity list
			if (idnew>=4)
			{
				// current cavity vector
				vector < int* >* vec;

				// first time this cavity is encountered
				if (delphi->cavitiesVec->size()<((unsigned)(idnew-4+1)))
				{
					//cout << endl << INFO << "Detected cavity " << idnew-4;
					vec = new vector<int*>();
					if (vec==NULL)
					{
						cout << endl << ERR << "Not enough memory to complete cavity detection, stopping";
						cout.flush();						
						exit(-1);
					}
					delphi->cavitiesVec->push_back(vec);
				}
				else
				{
					vec = delphi->cavitiesVec->at(idnew-4);
				}				
				int* v = allocateVector<int>(3);
				if (v==NULL)
				{
					cout << endl << ERR << "Not enough memory to complete cavity detection, stopping";
					cout.flush();
					exit(-1);
				}
				v[0] = cix;
				v[1] = ciy;
				v[2] = ciz;
				//printf("\n %d %d %d (%d,%d)",cix,ciy,ciz,idnew,idold);
				vec->push_back(v);
			}
		}
		else
			continue;
		
		if (cix+1<delphi->nx && (delphi->status[cix+1][ciy][ciz])==idold)
		{
			triplet.first = cix+1;
			triplet.second.first = ciy;
			triplet.second.second = ciz;
			myqueue.push(triplet);
		}
		if (ciy+1<delphi->ny && (delphi->status[cix][ciy+1][ciz])==idold)
		{
			triplet.first = cix;
			triplet.second.first = ciy+1;
			triplet.second.second = ciz;
			myqueue.push(triplet);		
		}		
		if (ciz+1<delphi->nz && (delphi->status[cix][ciy][ciz+1])==idold)
		{
			triplet.first = cix;
			triplet.second.first = ciy;
			triplet.second.second = ciz+1;
			myqueue.push(triplet);				
		}

		if (cix-1>=0 && (delphi->status[cix-1][ciy][ciz])==idold)
		{
			triplet.first = cix-1;
			triplet.second.first = ciy;
			triplet.second.second = ciz;
			myqueue.push(triplet);
		}

		if (ciy-1>=0 && (delphi->status[cix][ciy-1][ciz])==idold)
		{
			triplet.first = cix;
			triplet.second.first = ciy-1;
			triplet.second.second = ciz;
			myqueue.push(triplet);
		}

		if (ciz-1>=0 && (delphi->status[cix][ciy][ciz-1])==idold)
		{
			triplet.first = cix;
			triplet.second.first = ciy;
			triplet.second.second = ciz-1;
			myqueue.push(triplet);
		}
	}
	return;
}
 
bool Surface::getProjection(double p[3],double* proj1,double* proj2,
		double* proj3,double* normal1,double* normal2,double* normal3)
{
	cout << endl << WARN << "Projection not supported!";
	return false;
}

void Surface::getRayIntersection(double p1[3],double p2[3],vector<pair<double,double*> >& intersections,int thdID)
{
	cout << endl << WARN << "Ray-Patch intersection not supported!";
}

void Surface::updateSurf()
{
	cout << endl << WARN << "Surface update not supported!";
}


void Surface::projector(int start,int end)
{
	for (int i=start;i<end;i++)
	{
		double gridPoint[3];
		gridPoint[0] = delphi->x[delphi->ibgp[3*i]];
		gridPoint[1] = delphi->y[delphi->ibgp[3*i+1]];
		gridPoint[2] = delphi->z[delphi->ibgp[3*i+2]];
		getProjection(gridPoint,&(delphi->scspos[3*i]),
									&(delphi->scspos[3*i+1]),
									&(delphi->scspos[3*i+2]),
									&(delphi->scsnor[3*i]),
									&(delphi->scsnor[3*i+1]),
									&(delphi->scsnor[3*i+2]));
	}
}

void Surface::intersector(double* volPanel,int nb,int start,int end,int* numIntersections,int thdID)
{
	double diff[3];
	vector<pair < double, double* > > intersections;	
	intersections.reserve(2000);

	volPanel[panel]=0;
	double pa[3],pb[3];
	
	int NX = delphi->nx;
	int NY = delphi->ny;
	int NZ = delphi->nz;

#ifdef DEBUG_SURFACE
	// debug variables
	int** flag1,**flag2,**flag3;
	
	if (panel==0)
		flag1 = allocateMatrix2D<int>(end,nb);
	else if(panel==1)
		flag2 = allocateMatrix2D<int>(end,nb);
	else if(panel==2)
		flag3 = allocateMatrix2D<int>(end,nb);

	for (int i=start;i<end;i++)		
	{			
		for (int j=0;j<nb;j++)
		{
			if (panel==0)
				flag1[i][j]=0;
			else if (panel==1)
				flag2[i][j]=0;
			else
				flag3[i][j]=0;
		}
	}
#endif	

	double perturb = 0;

	// if cavities and espilon map is not necessary then skip this step at all
	if (!delphi->buildStatus && !delphi->buildEpsMap)
	{
		// skip
	}

	// if only cavity detection is required, only the first panel is ray traced
	else if (delphi->buildStatus && !delphi->buildEpsMap && panel!=0) 
	{
		// skip panel 1,2
	}

	// if both cavity detection and epsilon map are required, 3 panels must be traced
	else
	{

		for (int i=start;i<end;i++)
		{				
			int trials = 0;
			for (int j=0;j<nb;j++)
			{
				//printf("\n %d %d %d",i,j,trials);
				// light source and ray along coordinate direction
				if (panel==0)
				{
					pa[0] = delphi->x[0];
					pa[1] = delphi->y[i]+perturb;        
					pa[2] = delphi->z[j]+perturb;
					pb[0] = delphi->x[delphi->nx-1];
					pb[1] = pa[1];
					pb[2] = pa[2];
				}
				else if (panel==1)
				{										
					pa[0] = delphi->x[j]+perturb;
					pa[1] = delphi->y[i]+perturb;        
					pa[2] = delphi->z[0];
					pb[0] = pa[0];
					pb[1] = pa[1];
					pb[2] = delphi->z[delphi->nz-1];
				}
				else
				{
					pa[0] = delphi->x[j]+perturb;
					pa[2] = delphi->z[i]+perturb;
					pa[1] = delphi->y[0];   
					pb[0] = pa[0];	
					pb[1] = delphi->y[delphi->ny-1];
					pb[2] = pa[2];
				}
												
				SUB(diff,pb,pa);
			
				intersections.clear();

				getRayIntersection(pa,pb,intersections,thdID);

				if (intersections.size()==0)
					continue;
					
				// debug check wrong sequence. "Enter-Exit, Enter"
				/*intersections.clear();
				intersections.push_back(pair<double,double*>(3.0,(double*)1));
				intersections.push_back(pair<double,double*>(3.0,(double*)1));
				intersections.push_back(pair<double,double*>(4.0,(double*)1));
				intersections.push_back(pair<double,double*>(4.0,(double*)1));
				intersections.push_back(pair<double,double*>(5.0,(double*)1));
				intersections.push_back(pair<double,double*>(5.0,(double*)1));
				intersections.push_back(pair<double,double*>(5.0,(double*)1));*/

				// check if repetions allow the ray to enter and exist. If the
				// ray only enters then this is an error
				vector<pair<double,double*> >::iterator itt = intersections.begin();
				unsigned int vsize = intersections.size(),ind=0;
				bool closure = false;
				double lastValid = INFINITY;
				double a;
				
				for (;itt!=intersections.end();itt++)
				{
					a = (*itt).first;
					//printf("\n try to enter at -> %f",(*itt).first);
					// get the entering point
					//if (a!=lastValid)
					if (fabs(a-lastValid)>=EPS_INT)
					{
						//printf(" ok!");

						// ok the ray is entering
						closure = false;
						// search the exiting point of the ray
						while(1)
						{							
							ind++;							
							if (ind==vsize)
								break;
							itt++;
							
							//printf("\n try to exit at -> %f",(*itt).first);
							//if (a!=(*itt).first)
							if (fabs(a-(*itt).first)>=EPS_INT)
							{
								// ok the ray is exiting 

								//printf(" (ok!)");
								closure = true;
								lastValid = (*itt).first;
								break;	
							}
						}			
					}
					ind++;
				}

				/* single or multiple tangent intersection.*/
				if (lastValid==INFINITY)
				{
					//cout << " * ";				
					continue;
				}
				
				perturb = 0;
				// check finished.
				if (!closure)
				{
						// TEMP: go to next ray
						/*double debug_int[3],mydir[3];
						SUB(mydir,pb,pa)
						FILE* fp;
						fp = fopen("debug.txt","a");
						fprintf(fp,"\n//======================================");
						for (int uu=0;uu<intersections.size();uu++)
						{
							fprintf(fp,"\n // t value %f",intersections[uu]);
							ADD_MUL(debug_int,pa,mydir,(intersections[uu].first))											
							fprintf(fp,"\nsphere {"); 
							fprintf(fp,"\n<%f,%f,%f>,0.2\n",debug_int[0],debug_int[1],debug_int[2]);
							fprintf(fp,"\npigment{ color Yellow}}"); 	

						}
						fprintf(fp,"\ncylinder { <%f,%f,%f>,<%f,%f,%f>, 0.1 pigment{color Blue}}\n",pa[0],pa[1],pa[2],pb[0],pb[1],pb[2]);
						fprintf(fp,"\n//======================================");					
						fclose(fp);
						(*numIntersections)++;

						continue;
						*/

					// too many trials, cannot safely ray-trace, approximate the current ray with the previous one
					if (trials==100)
					{							
						if (panel==0)
						{
							if (delphi->buildStatus && delphi->buildEpsMap)
							{
								for (int ix=0;ix<delphi->nx;ix++)
								{
									delphi->EPSMAP((ix),(i),(j),0,NX,NY,NZ)= delphi->EPSMAP((ix),(i),(j-1),0,NX,NY,NZ);
									delphi->status[ix][i][j] = delphi->status[ix][i][j-1];
								}
							}
							else if (delphi->buildStatus)
							{
								for (int ix=0;ix<delphi->nx;ix++)
									delphi->status[ix][i][j] = delphi->status[ix][i][j-1];
							}
							// in this case idebmap is directly written because it is no more derived from status map
							else if (delphi->buildEpsMap)
							{
								for (int ix=0;ix<delphi->nx;ix++)						
								{
									delphi->EPSMAP((ix),(i),(j),0,NX,NY,NZ)= delphi->EPSMAP((ix),(i),(j-1),0,NX,NY,NZ);
									delphi->IDEBMAP(ix,i,j,(delphi->nx),(delphi->ny))=delphi->IDEBMAP(ix,i,j-1,(delphi->nx),(delphi->ny));
								}
							}
						}
						else if (panel==1)
						{
							//if (delphi->buildEpsMap)
								for (int iz=0;iz<delphi->nz;iz++)
									delphi->EPSMAP((j),(i),(iz),2,NX,NY,NZ)=delphi->EPSMAP((j-1),(i),(iz),2,NX,NY,NZ);
						}
						else if (panel==2)
						{
							//if (delphi->buildEpsMap)
								for (int iy=0;iy<delphi->ny;iy++)
									delphi->EPSMAP((j),(iy),(i),1,NX,NY,NZ)=delphi->EPSMAP((j-1),(iy),(i),1,NX,NY,NZ);	
						}
																
						continue;
					}
						
					if (trials==0)
						(*numIntersections)++;

					// tries with a little random perturbation of the ray direction
					perturb = 0.01*(2*(rand()/((float)RAND_MAX))-1);
					trials++;
					j--;
					
					#ifdef DEBUG_SURFACE
					if (panel==0)
						flag1[i][j]=-1;
					else if (panel==1)
						flag2[i][j]=-1;
					else if (panel==2)
						flag3[i][j]=-1;	
					#endif
					
					continue;				
				}
							
				#ifdef DEBUG_SURFACE
				if (panel==0)
					flag1[i][j]=+1;
				else if (panel==1)
					flag2[i][j]=+1;
				else if (panel==2)
					flag3[i][j]=+1;
				#endif
				
				// if there are intersections mark the epsmap cells that are inside
				// by counting the number of intersections. This is done for each panel
				// that is in the x,y,z directions respectively.
				// Eg. Tracing the ray one obtains inside/out information
				// 000000000|-1-1-1-1-1-1-1-1-1-1-1-1|000000000000|-1-1-1-1-1|000000			
				if (intersections.size()>1)
				{					
					if (panel==0)
					{
						double lim1,lim2;
						double lastValidLim2 = INFINITY;					

						for (unsigned int ig=0;ig<intersections.size();ig+=2)
						{										
							if ((ig+1)>=vsize)
								break;
							
							lim1 = pa[0]+diff[0]*intersections[ig].first;
							lim2 = pa[0]+diff[0]*intersections[ig+1].first;
						
							// skip repeated intersections 
							//if (lim2 == lim1 || lim1 == lastValidLim2)
							if (fabs(lim2-lim1)<EPS_INT || fabs(lim1-lastValidLim2)<EPS_INT)
							{																
								ig--;
								continue;
							}
							
							volPanel[panel]+=lim2-lim1;												
							lastValidLim2 = lim2;
													
							if (delphi->buildEpsMap && delphi->buildStatus)
							{
								for (int ix=0;ix<delphi->nx;ix++)
								{			
									//cout << endl;
									if (delphi->x[ix]+delphi->hside<=lim2 && delphi->x[ix]+delphi->hside>=lim1)
									{
										//delphi->epsmapx[ix][i][j]=-1;
										delphi->EPSMAP((ix),(i),(j),0,NX,NY,NZ)=inside;
										//cout << delphi->EPSMAP((ix),(i),(j),0,NX,NY,NZ);
									}

									if (delphi->x[ix]<=lim2 && delphi->x[ix]>=lim1)
										delphi->status[ix][i][j]=STAUTS_POINT_INSIDE;								
								}
							}
							else if (delphi->buildEpsMap)
							{
								for (int ix=0;ix<delphi->nx;ix++)
								{	
									// also idebmap is directly written because in this case status map is not used
									if (delphi->x[ix]+delphi->hside<=lim2 && delphi->x[ix]+delphi->hside>=lim1)
									{
										delphi->EPSMAP((ix),(i),(j),0,NX,NY,NZ)=inside;
										// mark inside
										delphi->IDEBMAP((ix),(i),(j),(delphi->nx),(delphi->ny))=false;
									}
								}
							}
							else if (delphi->buildStatus)
							{
								for (int ix=0;ix<delphi->nx;ix++)
									if (delphi->x[ix]<=lim2 && delphi->x[ix]>=lim1)
										delphi->status[ix][i][j]=STAUTS_POINT_INSIDE;
							}
						}
					}
					else if (panel==1)
					{
						double lim1,lim2;
						double lastValidLim2 = INFINITY;
						
						for (unsigned int ig=0;ig<intersections.size();ig+=2)
						{															
							if ((ig+1)>=vsize)
								break;
							
							lim1 = pa[2]+diff[2]*intersections[ig].first;
							lim2 = pa[2]+diff[2]*intersections[ig+1].first;						
					
							// skip repeated intersections 
							//if (lim2 == lim1 || lim1 == lastValidLim2)
							if (fabs(lim2-lim1)<EPS_INT || fabs(lim1-lastValidLim2)<EPS_INT)
							{
								ig--;
								continue;
							}
							
							volPanel[panel]+=lim2-lim1;
							lastValidLim2 = lim2;
					
							for (int iz=0;iz<delphi->nz;iz++)
							{						
								if (delphi->z[iz]+delphi->hside<=lim2 && delphi->z[iz]+delphi->hside>=lim1)
									//delphi->epsmapz[j][i][iz]=-1;								
									delphi->EPSMAP((j),(i),(iz),2,NX,NY,NZ)=inside;
							}
						}
					}
					else
					{
						double lim1,lim2;
						double lastValidLim2 = INFINITY;

						for (unsigned int ig=0;ig<intersections.size();ig+=2)
						{										
							if ((ig+1)>=vsize)
								break;						

							lim1 = pa[1]+diff[1]*intersections[ig].first;
							lim2 = pa[1]+diff[1]*intersections[ig+1].first;

							// skip repeated intersections 											
							//if (lim2 == lim1 || lim1 == lastValidLim2)
							if (fabs(lim2-lim1)<EPS_INT || fabs(lim1-lastValidLim2)<EPS_INT)
							{
								ig--;
								continue;
							}	
							
							volPanel[panel]+=lim2-lim1;									
							lastValidLim2 = lim2;
							
							for (int iy=0;iy<delphi->ny;iy++)
							{							
								if (delphi->y[iy]+delphi->hside<=lim2 && delphi->y[iy]+delphi->hside>=lim1)
									//delphi->epsmapy[j][iy][i]=-1;
									delphi->EPSMAP((j),(iy),(i),1,NX,NY,NZ)=inside;
							}
						}												
					}									
				}
			}
		}
	}
#ifdef DEBUG_SURFACE
	FILE* fp;
	if (panel==0)
	{
		fp = fopen("panel0.txt","w");
		for (int i=start;i<end;i++)		
		{			
			for (int j=0;j<nb;j++)
				fprintf(fp,"%d ",flag1[i][j]);
			fprintf(fp,"\n");
		}
	
	}
	else if (panel==1)
	{
		fp = fopen("panel1.txt","w");
		for (int i=start;i<end;i++)		
		{			
			for (int j=0;j<nb;j++)
				fprintf(fp,"%d ",flag2[i][j]);
			fprintf(fp,"\n");
		}	
	}
	else if (panel==2)
	{
		fp = fopen("panel2.txt","w");
		for (int i=start;i<end;i++)		
		{			
			for (int j=0;j<nb;j++)
				fprintf(fp,"%d ",flag3[i][j]);
			fprintf(fp,"\n");
		}	
	}
	fclose(fp);
#endif
	
	// if a triangulation is needed and we have not already a scalar field,
	// more rays need to be casted
	if (accurateTriangulation && !isAvailableScalarField)
	{
		double delta = delta_accurate_triangulation;
		double perturb = 0;
		for (int i=start;i<end;i++)
		{				
			int trials = 0;
			for (int j=0;j<nb;j++)
			{
				if (panel==0)
				{
					pa[0] = delphi->x[0]-delphi->hside+delta;
					pa[1] = delphi->y[i]-delphi->hside+delta+perturb;        
					pa[2] = delphi->z[j]-delphi->hside+delta+perturb;
					pb[0] = delphi->x[delphi->nx-1]-delphi->hside+delta;
					pb[1] = pa[1];
					pb[2] = pa[2];
				}
				else if (panel==1)
				{										
					pa[0] = delphi->x[j]-delphi->hside+delta+perturb;
					pa[1] = delphi->y[i]-delphi->hside+delta+perturb;        
					pa[2] = delphi->z[0]-delphi->hside+delta;
					pb[0] = pa[0];
					pb[1] = pa[1];
					pb[2] = delphi->z[delphi->nz-1]-delphi->hside+delta;
				}
				else
				{
					pa[0] = delphi->x[j]-delphi->hside+delta+perturb;
					pa[1] = delphi->y[0]-delphi->hside+delta;   
					pa[2] = delphi->z[i]-delphi->hside+delta+perturb;					
					pb[0] = pa[0];	
					pb[1] = delphi->y[delphi->ny-1]-delphi->hside+delta;
					pb[2] = pa[2];
				}
						
				SUB(diff,pb,pa);
				intersections.clear();
				getRayIntersection(pa,pb,intersections,thdID);

				if (intersections.size()==0)
					continue;

				vector<pair<double,double*> >::iterator itt = intersections.begin();
				unsigned int vsize = intersections.size(),ind=0;
				bool closure = false;
				double lastValid = INFINITY;
				double a;
				
				for (;itt!=intersections.end();itt++)
				{
					a = (*itt).first;
					//printf("\n try to enter at -> %f",(*itt).first);
					// get the entering point
					//if (a!=lastValid)
					if (fabs(a-lastValid)>=EPS_INT)
					{
						//printf(" ok!");

						// ok the ray is entering
						closure = false;
						// search the exiting point of the ray
						while(1)
						{							
							ind++;							
							if (ind==vsize)
								break;
							itt++;
							
							//printf("\n try to exit at -> %f",(*itt).first);
							//if (a!=(*itt).first)
							if (fabs(a-(*itt).first)>=EPS_INT)
							{
								// ok the ray is exiting 

								//printf(" (ok!)");
								closure = true;
								lastValid = (*itt).first;
								break;	
							}
						}			
					}
					ind++;
				}

				/* single or multiple tangent intersection.*/
				if (lastValid==INFINITY)
				{
					//cout << " * ";				
					continue;
				}
				
				perturb = 0;
				
				// check finished.
				if (!closure)
				{	
					/*
					// TEMP: go to next ray
					double debug_int[3],mydir[3];
					SUB(mydir,pb,pa)
					FILE* fp;
					
					fp = fopen("debug.txt","a");
					fprintf(fp,"\n//======================================");
					for (int uu=0;uu<intersections.size();uu++)
					{
						fprintf(fp,"\n // t value %f",intersections[uu]);
						ADD_MUL(debug_int,pa,mydir,(intersections[uu].first))											
						fprintf(fp,"\nsphere {"); 
						fprintf(fp,"\n<%f,%f,%f>,0.2\n",debug_int[0],debug_int[1],debug_int[2]);
						fprintf(fp,"\npigment{ color Yellow}}"); 	

					}
					fprintf(fp,"\ncylinder { <%f,%f,%f>,<%f,%f,%f>, 0.1 pigment{color Blue}}\n",pa[0],pa[1],pa[2],pb[0],pb[1],pb[2]);
					fprintf(fp,"\n//======================================");					
					fclose(fp);
					(*numIntersections)++;

					continue;
					*/

					// too many trials, cannot safely ray-trace, approximate the current ray with the previous one
					// analytical intersections cannot be recovered, only in/out position is recorded
					// The following marching cubes will be semi analytical. Semi means that were the analytical intersection
					// is not present the usual marching cubes rule will be used
					if (trials==100)
					{							
						if (panel==0)
							for (int ix=0;ix<delphi->nx;ix++)							
								verticesInsidenessMap[ix][i][j] = verticesInsidenessMap[ix][i][j-1];
						if (panel==1)
							for (int iz=0;iz<delphi->nz;iz++)							
								verticesInsidenessMap[j][i][iz] = verticesInsidenessMap[j-1][i][iz];
						if (panel==2)
							for (int iy=0;iy<delphi->ny;iy++)							
								verticesInsidenessMap[j][iy][i] = verticesInsidenessMap[j-1][iy][i];				

						
						//cout << endl << "pass here";
						
						continue;
					}					
										
					if (trials==0)
						(*numIntersections)++;

					// tries with a little random perturbation of the ray direction
					perturb = 0.01*(2*(rand()/((float)RAND_MAX))-1);
					trials++;
					j--;
					continue;
				}

				// if there are intersections mark the epsmap cells that are inside
				// by counting the number of intersections. This is done for each panel
				// that is in the x,y,z directions respectively.
				// Eg. Tracing the ray one obtains inside/out information
				// 000000000|-1-1-1-1-1-1-1-1-1-1-1-1|000000000000|-1-1-1-1-1|000000			
				if (intersections.size()>1)
				{					
					if (panel==0)
					{
						double lim1,lim2;
						double lastValidLim2 = INFINITY;					

						//printf("\n-----------");
						for (unsigned int ig=0;ig<intersections.size();ig+=2)
						{										
							if ((ig+1)>=vsize)
								break;
							
							lim1 = pa[0]+diff[0]*intersections[ig].first;
							lim2 = pa[0]+diff[0]*intersections[ig+1].first;
						
							// skip repeated intersections 
							//if (lim2 == lim1 || lim1 == lastValidLim2)
							if (fabs(lim2-lim1)<EPS_INT || fabs(lim1-lastValidLim2)<EPS_INT)
							{																
								ig--;
								continue;
							}
							
							lastValidLim2 = lim2;

							// if no cavity detection and no epsmap is enabled the volume must be computed here
							if (!delphi->buildStatus && !delphi->buildEpsMap)
								volPanel[panel]+=lim2-lim1;									

							double *intersec1,*intersec2;
							intersec1 = allocateVector<double>(3);
							intersec2 = allocateVector<double>(3);
							intersec1[0] = lim1;
							intersec1[1] = pa[1];
							intersec1[2] = pa[2];

							intersec2[0] = lim2;
							intersec2[1] = pa[1];
							intersec2[2] = pa[2];								
							
							// get the cube which the intersections belong. The little 1e-7 is 
							// used in order to be sure to get the right cube. Consider that
							// we are at the edge so by performing a little move on the right
							// direction we are sure to identify the right cube
							int ix_ = (int)rintp((intersec1[0]-delphi->xmin)/delphi->side);
							int iy_ = (int)rintp((intersec1[1]-delphi->ymin)/delphi->side);
							int iz_ = (int)rintp((intersec1[2]-delphi->zmin)/delphi->side);
							//int iy_ = i;
							//int iz_ = j;
							
							//printf("\n %d %d %d th %d",ix_,iy_,iz_,thdID);
							//getchar();

							// store the last intersection point (if the grid is accurate enough this is a unique point)
							if (intersectionsMatrixAlongX[ix_][iy_][iz_]!=NULL)
								deleteVector<double>(intersectionsMatrixAlongX[ix_][iy_][iz_]);

							intersectionsMatrixAlongX[ix_][iy_][iz_] = intersec1;

							ix_ = (int)rintp((intersec2[0]-delphi->xmin)/delphi->side);
							iy_ = (int)rintp((intersec2[1]-delphi->ymin)/delphi->side);
							iz_ = (int)rintp((intersec2[2]-delphi->zmin)/delphi->side);

							// store the last intersection point (if the grid is accurate enough this is a unique point)
							if (intersectionsMatrixAlongX[ix_][iy_][iz_]!=NULL)
								deleteVector<double>(intersectionsMatrixAlongX[ix_][iy_][iz_]);

							// store intersection point
							intersectionsMatrixAlongX[ix_][iy_][iz_] = intersec2;
							
							for (int ix=0;ix<delphi->nx;ix++)
							{			
								// in
								if (delphi->x[ix]-delphi->hside<=lim2  && delphi->x[ix]-delphi->hside>=lim1 )
									verticesInsidenessMap[ix][i][j] = false;
							}
						}
						//printf("\n-----------");
					}
					else if (panel==1)
					{
						double lim1,lim2;
						double lastValidLim2 = INFINITY;
						
						for (unsigned int ig=0;ig<intersections.size();ig+=2)
						{															
							if ((ig+1)>=vsize)
								break;
							
							lim1 = pa[2]+diff[2]*intersections[ig].first;
							lim2 = pa[2]+diff[2]*intersections[ig+1].first;						
					
							// skip repeated intersections 
							//if (lim2 == lim1 || lim1 == lastValidLim2)
							if (fabs(lim2-lim1)<EPS_INT || fabs(lim1-lastValidLim2)<EPS_INT)
							{
								ig--;
								continue;
							}
							
							
							lastValidLim2 = lim2;

							// if no cavity detection and no epsmap is enabled the volume must be computed here
							if (!delphi->buildStatus && !delphi->buildEpsMap)
								volPanel[panel]+=lim2-lim1;									

							double *intersec1,*intersec2;
							intersec1 = allocateVector<double>(3);
							intersec2 = allocateVector<double>(3);
							intersec1[0] = pa[0] ;
							intersec1[1] = pa[1] ;
							intersec1[2] = lim1 ;

							intersec2[0] = pa[0] ;
							intersec2[1] = pa[1] ;
							intersec2[2] = lim2 ;								
							
							// get the cube which the intersections belong. The little 1e-7 is 
							// used in order to be sure to get the right cube. Consider that
							// we are at the edge so by performing a little move on the right
							// direction we are sure to identify the right cube
							int ix_ = (int)rintp((intersec1[0]-delphi->xmin)/delphi->side);
							int iy_ = (int)rintp((intersec1[1]-delphi->ymin)/delphi->side);
							int iz_ = (int)rintp((intersec1[2]-delphi->zmin)/delphi->side);
							//int ix_ = j;
							//int iy_ = i; 

							if (intersectionsMatrixAlongZ[ix_][iy_][iz_]!=NULL)
								deleteVector<double>(intersectionsMatrixAlongZ[ix_][iy_][iz_]);

							// store intersection point
							intersectionsMatrixAlongZ[ix_][iy_][iz_] = intersec1;

							ix_ = (int)rintp((intersec2[0]-delphi->xmin)/delphi->side);
							iy_ = (int)rintp((intersec2[1]-delphi->ymin)/delphi->side);
							iz_ = (int)rintp((intersec2[2]-delphi->zmin)/delphi->side);

							if (intersectionsMatrixAlongZ[ix_][iy_][iz_]!=NULL)
								deleteVector<double>(intersectionsMatrixAlongZ[ix_][iy_][iz_]);

							// store intersection point
							intersectionsMatrixAlongZ[ix_][iy_][iz_] = intersec2;	

							for (int iz=0;iz<delphi->nz;iz++)
							{			
								// in
								if (delphi->z[iz]-delphi->hside<=lim2  && delphi->z[iz]-delphi->hside>=lim1 )
									verticesInsidenessMap[j][i][iz] = false;
							}
						}
					}
					else
					{
						double lim1,lim2;
						double lastValidLim2 = INFINITY;

						for (unsigned int ig=0;ig<intersections.size();ig+=2)
						{										
							if ((ig+1)>=vsize)
								break;						

							lim1 = pa[1]+diff[1]*intersections[ig].first;
							lim2 = pa[1]+diff[1]*intersections[ig+1].first;

							// skip repeated intersections 											
							//if (lim2 == lim1 || lim1 == lastValidLim2)
							if (fabs(lim2-lim1)<EPS_INT || fabs(lim1-lastValidLim2)<EPS_INT)
							{
								ig--;
								continue;
							}	
							
							lastValidLim2 = lim2;

							// if no cavity detection and no epsmap is enabled the volume must be computed here
							if (!delphi->buildStatus && !delphi->buildEpsMap)
								volPanel[panel]+=lim2-lim1;									

							double *intersec1,*intersec2;
							intersec1 = allocateVector<double>(3);
							intersec2 = allocateVector<double>(3);
							intersec1[0] = pa[0] ;
							intersec1[1] = lim1 ;
							intersec1[2] = pa[2] ;

							intersec2[0] = pa[0] ;
							intersec2[1] = lim2 ;
							intersec2[2] = pa[2];								
							
							// get the cube which the intersections belong. The little 1e-7 is 
							// used in order to be sure to get the right cube. Consider that
							// we are at the edge so by performing a little move on the right
							// direction we are sure to identify the right cube
							int ix_ = (int)rintp((intersec1[0]-delphi->xmin)/delphi->side);
							int iy_ = (int)rintp((intersec1[1]-delphi->ymin)/delphi->side);
							int iz_ = (int)rintp((intersec1[2]-delphi->zmin)/delphi->side);
							//int ix_ = j;
							//int iz_ = i;

							if (intersectionsMatrixAlongY[ix_][iy_][iz_]!=NULL)
								deleteVector<double>(intersectionsMatrixAlongY[ix_][iy_][iz_]);

							// store intersection point
							intersectionsMatrixAlongY[ix_][iy_][iz_] = intersec1;

							ix_ = (int)rintp((intersec2[0]-delphi->xmin)/delphi->side);
							iy_ = (int)rintp((intersec2[1]-delphi->ymin)/delphi->side);
							iz_ = (int)rintp((intersec2[2]-delphi->zmin)/delphi->side);

							if (intersectionsMatrixAlongY[ix_][iy_][iz_]!=NULL)
								deleteVector<double>(intersectionsMatrixAlongY[ix_][iy_][iz_]);

							// store intersection point
							intersectionsMatrixAlongY[ix_][iy_][iz_] = intersec2;					

							for (int iy=0;iy<delphi->ny;iy++)
							{			
								// in
								if (delphi->y[iy]-delphi->hside<=lim2  && delphi->y[iy]-delphi->hside>=lim1 )
									verticesInsidenessMap[j][iy][i] = false;
							}
						}												
					}									
				}				
			}
		}
	}

	// this debug works only in single threaded mode
	/*FILE* fp;
	fp = fopen("debug.txt","w");
	for (int i=0;i<NX;i++)
		for (int j=0;j<NY;j++)
			for (int k=0;k<NZ;k++)
			{
				if (intersectionsMatrixAlongY[i][j][k]!=NULL)
					fprintf(fp,"%lf %lf %lf\n",intersectionsMatrixAlongY[i][j][k][0],intersectionsMatrixAlongY[i][j][k][1],intersectionsMatrixAlongY[i][j][k][2]);
				if (intersectionsMatrixAlongX[i][j][k]!=NULL)
					fprintf(fp,"%lf %lf %lf\n",intersectionsMatrixAlongX[i][j][k][0],intersectionsMatrixAlongX[i][j][k][1],intersectionsMatrixAlongX[i][j][k][2]);
				if (intersectionsMatrixAlongZ[i][j][k]!=NULL)
					fprintf(fp,"%lf %lf %lf\n",intersectionsMatrixAlongZ[i][j][k][0],intersectionsMatrixAlongZ[i][j][k][1],intersectionsMatrixAlongZ[i][j][k][2]);
			}

	fclose(fp);

	fp = fopen("debug2.txt","w");
	for (int i=0;i<NX;i++)
		for (int j=0;j<NY;j++)
			for (int k=0;k<NZ;k++)
			{
				fprintf(fp, "%f %f %f %d\n",delphi->x[i]-delphi->hside,delphi->y[j]-delphi->hside,delphi->z[k]-delphi->hside,verticesInsidenessMap[i][j][k]);
			}

	fclose(fp);	
	*/
}  


void Surface::triangulateSurface(double isolevel,const char* fileName)
{
	int NX = delphi->nx;
	int NY = delphi->ny;
	int NZ = delphi->nz;

	double area = 0;

	double** positions = allocateMatrix2D<double>(8,3);
	int** triangles = allocateMatrix2D<int>(5,3);
	double** vertPointers = allocateVector<double*>(3);
	
	double* gx = delphi->x;
	double* gy = delphi->y;
	double* gz = delphi->z;
	double d_hside = delphi->hside;

	int** xedge = allocateMatrix2D<int>(NX,NY);
	int** yedge = allocateMatrix2D<int>(NX,NY);
	int** xedge_down = allocateMatrix2D<int>(NX,NY);
	int** yedge_down = allocateMatrix2D<int>(NX,NY);
	int** zedge = allocateMatrix2D<int>(NX,NY);
		
	//getchar();
	
	for (int k=0;k<NZ;k++)	
	{
		// first slice reset down slice indices
		if (k==0)
		{			
			for (int i=0;i<NX;i++)
				for (int j=0;j<NY;j++)
				{
					xedge_down[i][j]=-1;
					yedge_down[i][j]=-1;					
					xedge[i][j]=-1;
					yedge[i][j]=-1;
					zedge[i][j]=-1;
				}		
		}
		else
		{			
			int** temp = xedge_down;
			xedge_down = xedge;
			xedge = temp;

			temp = yedge_down;
			yedge_down = yedge;
			yedge = temp;

			// it should be not necessary
			for (int i=0;i<NX;i++)
				for (int j=0;j<NY;j++)
				{
					xedge[i][j]=-1;
					yedge[i][j]=-1;
					zedge[i][j]=-1;
				}
		}

		printf("\r%sMC Completion %.2f%%                             ",INFO,((double)k+1)/NZ*100.0);		
		for (int i=0;i<NX;i++)			
			for (int j=0;j<NY;j++)			
			{

				// skip the boundary of the grid, there will never be triangles here if
				// the grid is correctly built
				if (i==0 || i==(NX-1)|| j==0 || j==(NY-1) || k==0 || k==(NZ-1))				
					continue;

				double votes[8];

				// if an accurate description is available each inside/outside flag
				// and edge surface intersections were analytically computed by the ray-tracing routine
				// this allows to perform a marching cubes where each vertex of the mesh is granted to
				// to belong to the surface. 
				// This is an analytical marching cubes				
				if (accurateTriangulation && !isAvailableScalarField)
				{
					if (verticesInsidenessMap[i][j][k])
						votes[0] = +1;
					else
						votes[0] = -1;
					
					if (verticesInsidenessMap[i+1][j][k])
						votes[1] = +1;
					else
						votes[1] = -1;

					if (verticesInsidenessMap[i+1][j+1][k])
						votes[2] = +1;
					else
						votes[2] = -1;

					if (verticesInsidenessMap[i][j+1][k])
						votes[3] = +1;
					else
						votes[3] = -1;

					if (verticesInsidenessMap[i][j][k+1])
						votes[4] = +1;
					else
						votes[4] = -1;

					if (verticesInsidenessMap[i+1][j][k+1])
						votes[5] = +1;
					else
						votes[5] = -1;

					if (verticesInsidenessMap[i+1][j+1][k+1])
						votes[6] = +1;
					else
						votes[6] = -1;

					if (verticesInsidenessMap[i][j+1][k+1])
						votes[7] = +1;
					else
						votes[7] = -1;

					//for (int vertInd =0;vertInd<8;vertInd++)
					//	printf(" %f ",votes[vertInd]);

					//printf("\n");
					//getchar();
				}
				// Classical marching cube interpolating scalar field values
				else if (accurateTriangulation && isAvailableScalarField)
				{
					votes[0] = scalarField[i][j][k];
					votes[1] = scalarField[i+1][j][k];
					votes[2] = scalarField[i+1][j+1][k];
					votes[3] = scalarField[i][j+1][k];
					votes[4] = scalarField[i][j][k+1];
					votes[5] = scalarField[i+1][j][k+1];
					votes[6] = scalarField[i+1][j+1][k+1];
					votes[7] = scalarField[i][j+1][k+1];
				}
				
				int numTriangles;

				double xg = gx[i];
				double yg = gy[j];
				double zg = gz[k];

				double table[6];
				table[0] = xg - d_hside;
				table[1] = xg + d_hside;
				table[2] = yg - d_hside;
				table[3] = yg + d_hside;
				table[4] = zg - d_hside;
				table[5] = zg + d_hside;
				
				if (!accurateTriangulation)
					for (int vertInd =0;vertInd<8;vertInd++)
					{
						votes[vertInd] = getInsidness(i,j,k,vertInd);
						positions[vertInd][0]= table[mulTable[vertInd][0]];
						positions[vertInd][1]= table[mulTable[vertInd][1]];
						positions[vertInd][2]= table[mulTable[vertInd][2]];
					}
				else
					for (int vertInd =0;vertInd<8;vertInd++)
					{
						positions[vertInd][0]= table[mulTable[vertInd][0]];
						positions[vertInd][1]= table[mulTable[vertInd][1]];
						positions[vertInd][2]= table[mulTable[vertInd][2]];
					}
				
				// now that the votes and positions are obtained 
				// triangulate the cell by marching cube rule.
								
				numTriangles = getTriangles(votes,positions,isolevel,triangles,i,j,k,NX,NY,NZ,xedge,yedge,zedge,xedge_down,yedge_down);				
				
				//printf("\n Num tri %d",numTriangles);
				// save triangles
				for (int kk = 0;kk<numTriangles;kk++)
				{
					int tri_temp[3];					

					for (int ll = 0;ll<3;ll++)
					{										
						tri_temp[ll] = triangles[kk][ll];
						vertPointers[ll]=vertList[tri_temp[ll]];
					}
							
					double a =0,b=0,c=0;
					DIST(a,vertPointers[0],vertPointers[1]);
					DIST(b,vertPointers[0],vertPointers[2]);
					DIST(c,vertPointers[1],vertPointers[2]);
					double ttt = (a+b+c)*(b+c-a)*(c+a-b)*(a+b-c);
					if (ttt>0)						
						area += 0.25*sqrt(ttt);

					int* tri = allocateVector<int>(3);
					if (!isAvailableScalarField)
					{
						tri[2]=tri_temp[0];
						tri[1]=tri_temp[1];
						tri[0]=tri_temp[2];						
					}
					else
					{
						tri[0]=tri_temp[0];
						tri[1]=tri_temp[1];
						tri[2]=tri_temp[2];						
					}
					triList.push_back(tri);
				}				
				
			}
			
	}
	deleteMatrix2D<double>(8,positions);
	deleteMatrix2D<int>(5,triangles);
	deleteVector<double*>(vertPointers);

	deleteMatrix2D<int>(NX,xedge);
	deleteMatrix2D<int>(NX,yedge);
	deleteMatrix2D<int>(NX,zedge);
	deleteMatrix2D<int>(NX,xedge_down);
	deleteMatrix2D<int>(NX,yedge_down);
	
	totalSurfaceArea = area;

	cout << endl << INFO << "Total, grid conformant, surface area is " << area << " [A^2]";

	// save all in OFF format
	FILE* fp;
    fp = fopen(fileName,"w");

	int numVertexes = vertList.size();
	int numTriangles = triList.size();

	cout << endl << INFO << "Number of vertices " << numVertexes << " number of triangles " << numTriangles;
	cout << endl << INFO << "Writing triangulated surface in OFF file format in " << fileName << "...";

	if (fp==NULL)
	{
		cout << endl << WARN << "Cannot write file " << fileName;
		return;
	}
				
	fprintf(fp,"OFF\n");
	time_t pt;
    time(&pt);    
	fprintf(fp,"# File created by %s version %s date %s\n",PROGNAME,VERSION,ctime(&pt));	
	fprintf(fp,"%d %d 0\n",numVertexes,numTriangles);

	for (int i=0;i<numVertexes;i++)
		fprintf(fp,"%.3f %.3f %.3f\n",vertList[i][0],vertList[i][1],vertList[i][2]);
	
	for (int i=0;i<numTriangles;i++)
		fprintf(fp,"3 %d %d %d\n",triList[i][0],triList[i][1],triList[i][2]);
	
	fclose(fp);
	return;
}

void Surface::smoothSurface(const char* fn)
{
	#define MAX_NEIGHBOURS 20

	vector<double*> tempVertices;
	int nv = vertList.size();
	int nt = triList.size();
	tempVertices.resize(nv);

	for (int i=0;i<nv;i++)
		tempVertices[i]=allocateVector<double>(3);
	
	int *vertdeg[MAX_NEIGHBOURS];	
	bool flagvert;
	for(int i=0;i<MAX_NEIGHBOURS;i++)
		vertdeg[i]=allocateVector<int>(nv);		

	for(int i=0;i<nv;i++)
		vertdeg[0][i]=0;

	for(int i=0;i<nt;i++)
	{
		// first vertex
		flagvert=true;
		for(int j=0;j<vertdeg[0][triList[i][0]];j++)
		{
			if(triList[i][1]==vertdeg[j+1][triList[i][0]])
			{
				flagvert=false;
				break;
			}
		}
		if(flagvert)
		{
			vertdeg[0][triList[i][0]]++;
			vertdeg[vertdeg[0][triList[i][0]]][triList[i][0]]=triList[i][1];
			
		}
		flagvert=true;
		for(int j=0;j<vertdeg[0][triList[i][0]];j++)
		{
			if(triList[i][2]==vertdeg[j+1][triList[i][0]])
			{
				flagvert=false;
				break;
			}
		}
		if(flagvert)
		{
			vertdeg[0][triList[i][0]]++;
			vertdeg[vertdeg[0][triList[i][0]]][triList[i][0]]=triList[i][2];
			
		}
		//second vertex
		flagvert=true;
		for(int j=0;j<vertdeg[0][triList[i][1]];j++)
		{
			if(triList[i][0]==vertdeg[j+1][triList[i][1]])
			{
				flagvert=false;
				break;
			}
		}
		if(flagvert)
		{
			vertdeg[0][triList[i][1]]++;
			vertdeg[vertdeg[0][triList[i][1]]][triList[i][1]]=triList[i][0];
			
		}
		flagvert=true;
		for(int j=0;j<vertdeg[0][triList[i][1]];j++)
		{
			if(triList[i][2]==vertdeg[j+1][triList[i][1]])
			{
				flagvert=false;
				break;
			}
		}
		if(flagvert)
		{
			vertdeg[0][triList[i][1]]++;
			vertdeg[vertdeg[0][triList[i][1]]][triList[i][1]]=triList[i][2];
			
		}
		//third vertex
		flagvert=true;
		for(int j=0;j<vertdeg[0][triList[i][2]];j++)
		{
			if(triList[i][0]==vertdeg[j+1][triList[i][2]])
			{
				flagvert=false;
				break;
			}
		}
		if(flagvert)
		{
			vertdeg[0][triList[i][2]]++;
			vertdeg[vertdeg[0][triList[i][2]]][triList[i][2]]=triList[i][0];
			
		}
		flagvert=true;
		for(int j=0;j<vertdeg[0][triList[i][2]];j++)
		{
			if(triList[i][1]==vertdeg[j+1][triList[i][2]])
			{
				flagvert=false;
				break;
			}
		}
		if(flagvert)
		{
			vertdeg[0][triList[i][2]]++;
			vertdeg[vertdeg[0][triList[i][2]]][triList[i][2]]=triList[i][1];
			
		}
	}

	
	for(int i=0;i<nv;i++)
	{
		tempVertices[i][0]=0;
		tempVertices[i][1]=0;
		tempVertices[i][2]=0;
			
		int div = 0;		

		// forward
		// primary neighbours	
		for(int j=0;j<vertdeg[0][i];j++)
		{
			tempVertices[i][0]+=vertList[vertdeg[j+1][i]][0];
			tempVertices[i][1]+=vertList[vertdeg[j+1][i]][1];
			tempVertices[i][2]+=vertList[vertdeg[j+1][i]][2];
								
			div++;
		}
		tempVertices[i][0]/=(div);
		tempVertices[i][1]/=(div);
		tempVertices[i][2]/=(div);			

		tempVertices[i][0]=0.5*(vertList[i][0]+tempVertices[i][0]);
		tempVertices[i][1]=0.5*(vertList[i][1]+tempVertices[i][1]);
		tempVertices[i][2]=0.5*(vertList[i][2]+tempVertices[i][2]);
	}

	// save all
	for(int i=0;i<nv;i++)
	{
		vertList[i][0]=tempVertices[i][0];
		vertList[i][1]=tempVertices[i][1];
		vertList[i][2]=tempVertices[i][2];
	}	

	//delete all
	for (int i=0;i<nv;i++)
		deleteVector<double>(tempVertices[i]);

	for (int i=0;i<MAX_NEIGHBOURS;i++)
		deleteVector<int>(vertdeg[i]);

	// save all in OFF format
	FILE* fp;
    fp = fopen(fn,"w");
	
	if (fp==NULL)
	{
		cout << endl << WARN << "Cannot write file " << fn;
		return;
	}
				
	fprintf(fp,"OFF\n");
	time_t pt;
    time(&pt);    
	fprintf(fp,"# File created by %s version %s date %s\n",PROGNAME,VERSION,ctime(&pt));	
	fprintf(fp,"%d %d 0\n",nv,nt);

	for (int i=0;i<nv;i++)
		fprintf(fp,"%.3f %.3f %.3f\n",vertList[i][0],vertList[i][1],vertList[i][2]);
	
	for (int i=0;i<nt;i++)
		fprintf(fp,"3 %d %d %d\n",triList[i][0],triList[i][1],triList[i][2]);
	
	fclose(fp);
	return;
}

char Surface::getInsidness(int i, int j, int k, int vertInd)
{
	short insideness[8];
	int NX = delphi->nx;
	int NY = delphi->ny;
	int NZ = delphi->nz;

	if (vertInd == 3)
	{
		insideness[0]=delphi->status[i][j][k];
		// saturate outvalue to +1 such that cavities or temporary out points are all the same
		if (insideness[0]>0)
			insideness[0]=+1;

		if (i-1>=0)
			insideness[1]=delphi->status[i-1][j][k];
		else
			insideness[1]=+1;
		
		if (insideness[1]>0)
			insideness[1]=+1;
	
		if (i-1>=0 && k-1>=0)
			insideness[2]=delphi->status[i-1][j][k-1];
		else
			insideness[2]=+1;

		if (insideness[2]>0)
			insideness[2]=+1;


		if (k-1>=0)
			insideness[3]=delphi->status[i][j][k-1];
		else
			insideness[3]=+1;

		if (insideness[3]>0)
			insideness[3]=+1;

		if (j+1<NY)
			insideness[4]=delphi->status[i][j+1][k];
		else
			insideness[4]=+1;

		if (insideness[4]>0)
			insideness[4]=+1;

		if (j+1<NY && i-1>=0)
			insideness[5]=delphi->status[i-1][j+1][k];
		else
			insideness[5]=+1;

		if (insideness[5]>0)
			insideness[5]=+1;

		if (j+1<NY && i-1>=0 && k-1>=0)
			insideness[6]=delphi->status[i-1][j+1][k-1];
		else
			insideness[6]=+1;

		if (insideness[6]>0)
			insideness[6]=+1;

		if (j+1<NY && k-1>=0)
			insideness[7]=delphi->status[i][j+1][k-1];
		else
			insideness[7]=+1;

		if (insideness[7]>0)
			insideness[7]=+1;

	}
	else if (vertInd==2)
	{
		insideness[0]=delphi->status[i][j][k];
		// saturate outvalue to +1 such that cavities or temporary out points are all the same
		if (insideness[0]>0)
			insideness[0]=+1;

		if (i+1<NX)
			insideness[1]=delphi->status[i+1][j][k];
		else
			insideness[1]=+1;
		
		if (insideness[1]>0)
			insideness[1]=+1;
	
		if (i+1<NX && j+1<NY)
			insideness[2]=delphi->status[i+1][j+1][k];
		else
			insideness[2]=+1;

		if (insideness[2]>0)
			insideness[2]=+1;


		if (j+1<NY)
			insideness[3]=delphi->status[i][j+1][k];
		else
			insideness[3]=+1;

		if (insideness[3]>0)
			insideness[3]=+1;

		if (k-1>=0)
			insideness[4]=delphi->status[i][j][k-1];
		else
			insideness[4]=+1;

		if (insideness[4]>0)
			insideness[4]=+1;

		if (k-1>=0 && i+1<NX)
			insideness[5]=delphi->status[i+1][j][k-1];
		else
			insideness[5]=+1;

		if (insideness[5]>0)
			insideness[5]=+1;

		if (k-1>=0 && i+1<NX && j+1<NY)
			insideness[6]=delphi->status[i+1][j+1][k-1];
		else
			insideness[6]=+1;

		if (insideness[6]>0)
			insideness[6]=+1;

		if (k-1>=0 && j+1<NY)
			insideness[7]=delphi->status[i][j+1][k-1];
		else
			insideness[7]=+1;

		if (insideness[7]>0)
			insideness[7]=+1;
	}
	else if (vertInd==1)
	{
		insideness[0]=delphi->status[i][j][k];
		// saturate outvalue to +1 such that cavities or temporary out points are all the same
		if (insideness[0]>0)
			insideness[0]=+1;

		if (i+1<NX)
			insideness[1]=delphi->status[i+1][j][k];
		else
			insideness[1]=+1;
		
		if (insideness[1]>0)
			insideness[1]=+1;
	
		if (i+1<NX && j-1>=0)
			insideness[2]=delphi->status[i+1][j-1][k];
		else
			insideness[2]=+1;

		if (insideness[2]>0)
			insideness[2]=+1;


		if (j-1>=0)
			insideness[3]=delphi->status[i][j-1][k];
		else
			insideness[3]=+1;

		if (insideness[3]>0)
			insideness[3]=+1;

		if (k-1>=0)
			insideness[4]=delphi->status[i][j][k-1];
		else
			insideness[4]=+1;

		if (insideness[4]>0)
			insideness[4]=+1;

		if (k-1>=0 && i+1<NX)
			insideness[5]=delphi->status[i+1][j][k-1];
		else
			insideness[5]=+1;

		if (insideness[5]>0)
			insideness[5]=+1;

		if (k-1>=0 && i+1<NX && j-1>=0)
			insideness[6]=delphi->status[i+1][j-1][k-1];
		else
			insideness[6]=+1;

		if (insideness[6]>0)
			insideness[6]=+1;

		if (k-1>=0 && j-1>=0)
			insideness[7]=delphi->status[i][j-1][k-1];
		else
			insideness[7]=+1;

		if (insideness[7]>0)
			insideness[7]=+1;	
	}
	else if (vertInd==0)
	{
		insideness[0]=delphi->status[i][j][k];
		// saturate outvalue to +1 such that cavities or temporary out points are all the same
		if (insideness[0]>0)
			insideness[0]=+1;

		if (i-1>=0)
			insideness[1]=delphi->status[i-1][j][k];
		else
			insideness[1]=+1;
		
		if (insideness[1]>0)
			insideness[1]=+1;
	
		if (i-1>=0 && k-1>=0)
			insideness[2]=delphi->status[i-1][j][k-1];
		else
			insideness[2]=+1;

		if (insideness[2]>0)
			insideness[2]=+1;


		if (k-1>=0)
			insideness[3]=delphi->status[i][j][k-1];
		else
			insideness[3]=+1;

		if (insideness[3]>0)
			insideness[3]=+1;

		if (j-1>=0)
			insideness[4]=delphi->status[i][j-1][k];
		else
			insideness[4]=+1;

		if (insideness[4]>0)
			insideness[4]=+1;

		if (j-1>=0 && i-1>=0)
			insideness[5]=delphi->status[i-1][j-1][k];
		else
			insideness[5]=+1;

		if (insideness[5]>0)
			insideness[5]=+1;

		if (j-1>=0 && i-1>=0 && k-1>=0)
			insideness[6]=delphi->status[i-1][j-1][k-1];
		else
			insideness[6]=+1;

		if (insideness[6]>0)
			insideness[6]=+1;

		if (j-1>=0 && k-1>=0)
			insideness[7]=delphi->status[i][j-1][k-1];
		else
			insideness[7]=+1;

		if (insideness[7]>0)
			insideness[7]=+1;
	
	}
	else if (vertInd==7)
	{
		insideness[0]=delphi->status[i][j][k];
		// saturate outvalue to +1 such that cavities or temporary out points are all the same
		if (insideness[0]>0)
			insideness[0]=+1;

		if (i-1>=0)
			insideness[1]=delphi->status[i-1][j][k];
		else
			insideness[1]=+1;
		
		if (insideness[1]>0)
			insideness[1]=+1;
	
		if (i-1>=0 && k+1<NZ)
			insideness[2]=delphi->status[i-1][j][k+1];
		else
			insideness[2]=+1;

		if (insideness[2]>0)
			insideness[2]=+1;


		if (k+1<NZ)
			insideness[3]=delphi->status[i][j][k+1];
		else
			insideness[3]=+1;

		if (insideness[3]>0)
			insideness[3]=+1;

		if (j+1<NY)
			insideness[4]=delphi->status[i][j+1][k];
		else
			insideness[4]=+1;

		if (insideness[4]>0)
			insideness[4]=+1;

		if (j+1<NY && i-1>=0)
			insideness[5]=delphi->status[i-1][j+1][k];
		else
			insideness[5]=+1;

		if (insideness[5]>0)
			insideness[5]=+1;

		if (j+1<NY && i-1>=0 && k+1<NZ)
			insideness[6]=delphi->status[i-1][j+1][k+1];
		else
			insideness[6]=+1;

		if (insideness[6]>0)
			insideness[6]=+1;

		if (j+1<NY && k+1<NZ)
			insideness[7]=delphi->status[i][j+1][k+1];
		else
			insideness[7]=+1;

		if (insideness[7]>0)
			insideness[7]=+1;

	}
	else if (vertInd==6)
	{
		insideness[0]=delphi->status[i][j][k];
		// saturate outvalue to +1 such that cavities or temporary out points are all the same
		if (insideness[0]>0)
			insideness[0]=+1;

		if (i+1<NX)
			insideness[1]=delphi->status[i+1][j][k];
		else
			insideness[1]=+1;
		
		if (insideness[1]>0)
			insideness[1]=+1;
	
		if (i+1<NX && j+1<NY)
			insideness[2]=delphi->status[i+1][j+1][k];
		else
			insideness[2]=+1;

		if (insideness[2]>0)
			insideness[2]=+1;


		if (j+1<NY)
			insideness[3]=delphi->status[i][j+1][k];
		else
			insideness[3]=+1;

		if (insideness[3]>0)
			insideness[3]=+1;

		if (k+1<NZ)
			insideness[4]=delphi->status[i][j][k+1];
		else
			insideness[4]=+1;

		if (insideness[4]>0)
			insideness[4]=+1;

		if (k+1<NZ && i+1<NX)
			insideness[5]=delphi->status[i+1][j][k+1];
		else
			insideness[5]=+1;

		if (insideness[5]>0)
			insideness[5]=+1;

		if (k+1<NZ && i+1<NX && j+1<NY)
			insideness[6]=delphi->status[i+1][j+1][k+1];
		else
			insideness[6]=+1;

		if (insideness[6]>0)
			insideness[6]=+1;

		if (k+1<NZ && j+1<NY)
			insideness[7]=delphi->status[i][j+1][k+1];
		else
			insideness[7]=+1;

		if (insideness[7]>0)
			insideness[7]=+1;
	
	}
	else if (vertInd==5)
	{
		insideness[0]=delphi->status[i][j][k];
		// saturate outvalue to +1 such that cavities or temporary out points are all the same
		if (insideness[0]>0)
			insideness[0]=+1;

		if (i+1<NX)
			insideness[1]=delphi->status[i+1][j][k];
		else
			insideness[1]=+1;
		
		if (insideness[1]>0)
			insideness[1]=+1;
	
		if (i+1<NX && j-1>=0)
			insideness[2]=delphi->status[i+1][j-1][k];
		else
			insideness[2]=+1;

		if (insideness[2]>0)
			insideness[2]=+1;


		if (j-1>=0)
			insideness[3]=delphi->status[i][j-1][k];
		else
			insideness[3]=+1;

		if (insideness[3]>0)
			insideness[3]=+1;

		if (k+1<NZ)
			insideness[4]=delphi->status[i][j][k+1];
		else
			insideness[4]=+1;

		if (insideness[4]>0)
			insideness[4]=+1;

		if (k+1<NZ && i+1<NX)
			insideness[5]=delphi->status[i+1][j][k+1];
		else
			insideness[5]=+1;

		if (insideness[5]>0)
			insideness[5]=+1;

		if (k+1<NZ && i+1<NX && j-1>=0)
			insideness[6]=delphi->status[i+1][j-1][k+1];
		else
			insideness[6]=+1;

		if (insideness[6]>0)
			insideness[6]=+1;

		if (k+1<NZ && j-1>=0)
			insideness[7]=delphi->status[i][j-1][k+1];
		else
			insideness[7]=+1;

		if (insideness[7]>0)
			insideness[7]=+1;		
	}
	else if (vertInd==4)
	{
		insideness[0]=delphi->status[i][j][k];
		// saturate outvalue to +1 such that cavities or temporary out points are all the same
		if (insideness[0]>0)
			insideness[0]=+1;

		if (i-1>=0)
			insideness[1]=delphi->status[i-1][j][k];
		else
			insideness[1]=+1;
		
		if (insideness[1]>0)
			insideness[1]=+1;
	
		if (i-1>=0 && k+1<NZ)
			insideness[2]=delphi->status[i-1][j][k+1];
		else
			insideness[2]=+1;

		if (insideness[2]>0)
			insideness[2]=+1;


		if (k+1<NZ)
			insideness[3]=delphi->status[i][j][k+1];
		else
			insideness[3]=+1;

		if (insideness[3]>0)
			insideness[3]=+1;

		if (j-1>=0)
			insideness[4]=delphi->status[i][j-1][k];
		else
			insideness[4]=+1;

		if (insideness[4]>0)
			insideness[4]=+1;

		if (j-1>=0 && i-1>=0)
			insideness[5]=delphi->status[i-1][j-1][k];
		else
			insideness[5]=+1;

		if (insideness[5]>0)
			insideness[5]=+1;

		if (j-1>=0 && i-1>=0 && k+1<NZ)
			insideness[6]=delphi->status[i-1][j-1][k+1];
		else
			insideness[6]=+1;

		if (insideness[6]>0)
			insideness[6]=+1;

		if (j-1>=0 && k+1<NZ)
			insideness[7]=delphi->status[i][j-1][k+1];
		else
			insideness[7]=+1;

		if (insideness[7]>0)
			insideness[7]=+1;
	
	}
	else
	{
		cout << endl << ERR << "Unknown vertex index!";
		return false;
	}
	
	int acc = 0;
	// decide insidness
	for (int i=0;i<8;i++)
	{
		acc += (int)insideness[i];
	}
	// positively biased decision
	if (acc>0)
		return +1;
	else 
		return -1;
}


void Surface::vertexInterp(double isolevel, double* p1, double* p2, double valp1, double valp2,double* p) 
{
   double mu;
   mu = (isolevel - valp1) / (valp2 - valp1);
   double temp[3];
   SUB(temp,p2,p1);
   ADD_MUL(p,p1,temp,mu);
}


/**
   Given a grid cell and an isolevel, calculate the triangular
   facets required to represent the isosurface through the cell.
   Return the number of triangular facets, the matrix triangles
   will be loaded up with the vertices at most 5 triangular facets.
	0 will be returned if the grid cell is either totally above
   of totally below the isolevel. Can act using interpolation or querying
   the analytically computed intersection point. In this last case it is garanted
   that every point of the mesh belongs to the surface; ix,iy,iz are passed in order to query
   the intersection matrices
*/
int Surface::getTriangles(double* vertexValues,double** vertexPos,double isolevel, int** triangles,int ix,int iy,
						  int iz,int NX, int NY,int NZ,int** xedge,int** yedge,int** zedge,int** xedge_down,int** yedge_down)
{
	// Marching cubes vertex indices and edges convention; vertInd is indicated in the cube
	//
	//		   v4_________e4_________v5
	//			/|  				/|
	//		e7 / |                 / |
	//		  /  |             e5 /  |
	//		 /   | e8            /	 | e9
	//	  v7/____|_____e6_______/v6	 |
	//		|	 |              |	 |
	//	 	|  v0|______e0______|____|v1
	//	e11 |	/        		|   / 	
	//		|  /			e10	|  / 
	//		| /	e3				| / e1
	//		|/					|/
	//	  v3/_________e2________/v2
	//            
	int ntriang;
	int cubeindex;
	bool isApproximated[12] = {false,false,false,false,false,false,false,false,false,false,false,false};

	int vertlist_indexes[12]={-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
		
	double* vertlist[12];
	double vertlist_p[12][3];

	// classical mc
	if ((!accurateTriangulation) || (accurateTriangulation && isAvailableScalarField))
		for (int i=0;i<12;i++)
			vertlist[i]=vertlist_p[i];   
	// we already have the vertices, we don't need interpolation
	else
		for (int i=0;i<12;i++)
			vertlist[i]=NULL;
		
	//   Determine the index into the edge table which
	//   tells us which vertices are inside of the surface   
	cubeindex = 0;
	if (vertexValues[0] < isolevel) cubeindex |= 1;
	if (vertexValues[1] < isolevel) cubeindex |= 2;
	if (vertexValues[2] < isolevel) cubeindex |= 4;
	if (vertexValues[3] < isolevel) cubeindex |= 8;
	if (vertexValues[4] < isolevel) cubeindex |= 16;
	if (vertexValues[5] < isolevel) cubeindex |= 32;
	if (vertexValues[6] < isolevel) cubeindex |= 64;
	if (vertexValues[7] < isolevel) cubeindex |= 128;

	// Cube is entirely in/out of the surface 
	if (edgeTable[cubeindex] == 0)
		return 0;
   
	int vert_offset,i=ix,j=iy;

	// Find the vertices where the surface intersects the cube (classical)
	if ((!accurateTriangulation) || (accurateTriangulation && isAvailableScalarField))
	{		
		// 0 buffered
		if (edgeTable[cubeindex] & 1)	
		{
			vert_offset = xedge_down[i][j-1];
			vertlist_indexes[0] = vert_offset;
		}
		// 1 buffered
		if (edgeTable[cubeindex] & 2)
		{
			vert_offset = yedge_down[i][j];
			vertlist_indexes[1] = vert_offset;
		}
		// 2 buffered
		if (edgeTable[cubeindex] & 4)
		{
			vert_offset = xedge_down[i][j];
			vertlist_indexes[2] = vert_offset;
		}
		// 3 buffered
		if (edgeTable[cubeindex] & 8)
		{
			vert_offset = yedge_down[i-1][j];
			vertlist_indexes[3] = vert_offset;
		}
		// 4 buffered
		if (edgeTable[cubeindex] & 16)
		{
			vert_offset = xedge[i][j-1];
			vertlist_indexes[4] = vert_offset;
		}
		// 5 
		if (edgeTable[cubeindex] & 32)
		{
			vertexInterp(isolevel,vertexPos[5],vertexPos[6],vertexValues[5],vertexValues[6],vertlist[5]);
			double* vert = allocateVector<double>(3);
			ASSIGN(vert,vertlist[5])
			vertList.push_back(vert);
			vertlist_indexes[5] = vertList.size()-1;
			yedge[i][j] = vertList.size()-1;
		}
		// 6
		if (edgeTable[cubeindex] & 64)	
		{
			vertexInterp(isolevel,vertexPos[6],vertexPos[7],vertexValues[6],vertexValues[7],vertlist[6]);
			double* vert = allocateVector<double>(3);
			ASSIGN(vert,vertlist[6])
			vertList.push_back(vert);
			vertlist_indexes[6] = vertList.size()-1;
			vert_offset = vertList.size()-1;
			xedge[i][j] = vert_offset;
		}
		// 7 buffered
		if (edgeTable[cubeindex] & 128)
		{
			vert_offset = yedge[i-1][j];
			vertlist_indexes[7] = vert_offset;
		}
		// 8 buffered
		if (edgeTable[cubeindex] & 256)
		{
			vert_offset = zedge[i-1][j-1];			
			vertlist_indexes[8] = vert_offset;
		}
		// 9 buffered
		if (edgeTable[cubeindex] & 512)
		{
			vert_offset = zedge[i][j-1];
			vertlist_indexes[9] = vert_offset;
		}
		// 10
		if (edgeTable[cubeindex] & 1024) 
		{
			vertexInterp(isolevel,vertexPos[2],vertexPos[6],vertexValues[2],vertexValues[6],vertlist[10]);
			double* vert = allocateVector<double>(3);
			ASSIGN(vert,vertlist[10])
			vertList.push_back(vert);
			vert_offset = vertList.size()-1;
			zedge[i][j]=vert_offset;
			vertlist_indexes[10] = vert_offset;
		}
		// 11 buffered
		if (edgeTable[cubeindex] & 2048)
		{ 
			vert_offset = zedge[i-1][j];
			vertlist_indexes[11] = vert_offset;
		}
	}
	// the vertices are already analytically computed, they are only recovered. If there is an inconsistency
	// due to cavities removal or other numerical instabilities then some vertices are approximated by usual mc rule. The incidence of approximated
	// vertices is a non trivial function of the number of cavities, of their shape and their closeness to the
	// true surface
	else
	{	
		// 0 buffered
		if (edgeTable[cubeindex] & 1)	
		{
			vert_offset = xedge_down[i][j-1];
			vertlist_indexes[0] = vert_offset;
		}
		// 1 buffered
		if (edgeTable[cubeindex] & 2)
		{
			vert_offset = yedge_down[i][j];
			vertlist_indexes[1] = vert_offset;
		}
		// 2 buffered
		if (edgeTable[cubeindex] & 4)
		{
			vert_offset = xedge_down[i][j];
			vertlist_indexes[2] = vert_offset;
		}
		// 3 buffered
		if (edgeTable[cubeindex] & 8)
		{
			vert_offset = yedge_down[i-1][j];
			vertlist_indexes[3] = vert_offset;
		}
		// 4 buffered
		if (edgeTable[cubeindex] & 16)
		{
			vert_offset = xedge[i][j-1];
			vertlist_indexes[4] = vert_offset;
		}

		// 5
		if (edgeTable[cubeindex] & 32)
		{
			vertlist[5]	= intersectionsMatrixAlongY[ix+1][iy][iz+1];						
			// no vertex?! cavity detector generated this, generate new vertex
			if (vertlist[5]==NULL)
			{
				isApproximated[5]=true;			
				vertlist[5]=vertlist_p[5];
				vertexInterp(isolevel,vertexPos[5],vertexPos[6],vertexValues[5],vertexValues[6],vertlist[5]);
			}
			double* vert = allocateVector<double>(3);
			ASSIGN(vert,vertlist[5])
			vertList.push_back(vert);
			vertlist_indexes[5] = vertList.size()-1;
			yedge[i][j] = vertList.size()-1;
		}

		// 6
		if (edgeTable[cubeindex] & 64)
		{
			vertlist[6]	=	intersectionsMatrixAlongX[ix][iy+1][iz+1];			
			// no vertex?! cavity detector generated this, generate new vertex
			if (vertlist[6]==NULL)
			{
				//printf("\ndoh!");
				isApproximated[6]=true;			
				vertlist[6]=vertlist_p[6];
				vertexInterp(isolevel,vertexPos[6],vertexPos[7],vertexValues[6],vertexValues[7],vertlist[6]);
			}
			double* vert = allocateVector<double>(3);
			ASSIGN(vert,vertlist[6])
			vertList.push_back(vert);
			vertlist_indexes[6] = vertList.size()-1;
			vert_offset = vertList.size()-1;
			xedge[i][j] = vert_offset;
		}

		// 7 buffered
		if (edgeTable[cubeindex] & 128)
		{
			vert_offset = yedge[i-1][j];
			vertlist_indexes[7] = vert_offset;
		}
		// 8 buffered
		if (edgeTable[cubeindex] & 256)
		{
			vert_offset = zedge[i-1][j-1];			
			vertlist_indexes[8] = vert_offset;
		}
		// 9 buffered
		if (edgeTable[cubeindex] & 512)
		{
			vert_offset = zedge[i][j-1];
			vertlist_indexes[9] = vert_offset;
		}
		// 10
		if (edgeTable[cubeindex] & 1024)
		{
			vertlist[10]=	intersectionsMatrixAlongZ[ix+1][iy+1][iz];			
			// no vertex?! cavity detector generated this, generate new vertex
			if (vertlist[10]==NULL)
			{
				isApproximated[10]=true;
				vertlist[10]=vertlist_p[10];
				vertexInterp(isolevel,vertexPos[2],vertexPos[6],vertexValues[2],vertexValues[6],vertlist[10]);
			}
			double* vert = allocateVector<double>(3);			
			ASSIGN(vert,vertlist[10])
			vertList.push_back(vert);
			vert_offset = vertList.size()-1;
			zedge[i][j]=vert_offset;
			vertlist_indexes[10] = vert_offset;			
		}

		// 11 buffered
		if (edgeTable[cubeindex] & 2048)
		{ 
			vert_offset = zedge[i-1][j];
			vertlist_indexes[11] = vert_offset;
		}
	}

   // Create the triangles
   ntriang = 0;
   for (i=0;triTable[cubeindex][i]!=-1;i+=3) 
   {
      if (vertlist_indexes[triTable[cubeindex][i  ]]==-1)
	  {		  
		  cout << endl << WARN << "Mesh with hole!";
		  continue;
	  }
	  if (vertlist_indexes[triTable[cubeindex][i+1  ]]==-1)
	  {
		  cout << endl << WARN << "Mesh with hole!";
		  continue;
	  }
	  if (vertlist_indexes[triTable[cubeindex][i+2  ]]==-1)
	  {
		  cout << endl << WARN << "Mesh with hole!";
		  continue;
	  }

	  // save all the assigned indexes
	  triangles[ntriang][0] = vertlist_indexes[triTable[cubeindex][i  ]];
	  triangles[ntriang][1] = vertlist_indexes[triTable[cubeindex][i+1]];
	  triangles[ntriang][2] = vertlist_indexes[triTable[cubeindex][i+2]];	  
      ntriang++;
   }
   return ntriang;
}